#!/usr/bin/python
# -*- coding: utf-8 -*-

# Function for creating stimulus for Rasmus Eklund's experiment)
# Created: MN-2017-10-10; Revised: RE-2018-10-25

import mn, random
import numpy as np

# AAN defs --------------------------------------------------------------------#
def calc_aware(sub_list):
    '''
    Keyword arguments:
    sub_list -- list of numbers: 0 = unaware, 1 = aware

    Return:
    aware -- (aware/(aware + unaware)) (float)
    '''

    aware_index = []
    # find and enumerate all reported aware (sub_list)
    for i, x in enumerate(sub_list):
        if x == 1:
            aware_index += [i]

    # calculate percent reported aware
    aware = sum(sub_list)*1.0/len(sub_list)

    return aware

def calc(obj_list, sub_list):
    '''
    Calculates percentage of trials:
        reported as aware, correct, aware and correct, unaware and correct
    
    Keyword arguments:
    obj_list -- list of numbers: 0 = incorrect, 1 = correct
    sub_list -- list of numbers: 0 = unaware, 1 = aware
    
    Return:
    aware -- (aware/(aware + unaware)) (float)
    correct -- correct/(correct + incorrect))
    aware_correct -- (correct aware/(correct aware + correct unaware))
    unaware_correct -- (correct unaware/(correct aware + correct unaware))
    '''
    aware_index = []
    unaware_index = []
    aware_corr_index = []
    unaware_corr_index = []

    # find and enumerate all reported aware (sub_list)
    for i, x in enumerate(sub_list):
        if x == 1:
            aware_index += [i]

    # find the corresponding correct and incorrect responses (obj_list)
    for i in aware_index:
        aware_corr_index += [obj_list[i]]

    # find and enumerate all reported unaware (sub_list)
    for i, x in enumerate(sub_list):
        if x == 0:
            unaware_index += [i]

    # find the corresponding correct and incorrect responses (obj_list)
    for i in unaware_index:
        unaware_corr_index += [obj_list[i]]

    # calculate percent reported aware
    aware = sum(sub_list)*1.0/len(sub_list)
    # calculate percent correct
    correct = sum(obj_list)*1.0/len(obj_list)
    # calulate percent correct when reported aware
    if sum(aware_corr_index) != 0:
        aware_correct = sum(aware_corr_index)*1.0/len(aware_corr_index)
    else:
        aware_correct = 0
    # calculate percent correct when reported not aware
    if sum(unaware_corr_index) != 0:
        unaware_correct = sum(unaware_corr_index)*1.0/len(unaware_corr_index)
    else:
        unaware_correct = 0
    return aware, correct, aware_correct, unaware_correct
    
def create_stimulus(condition = 1, frequency = 1000, sdur = 0.1, s_lvl = 0.0, fade = 0.005, fs = 48000):
    ''' Creates stimulus for Rasmus experiment (add more info here.....)
    
    Keyword arguments:
    condition -- 1 = tone, 2 = catch
    stype -- Type of signal: 1 = sinusoid, 2 = square pulse (int)
    sdur -- duration of signal in seconds (float)
    s_lvl -- Level of signal relative to whote noise, rms dB (float)
    fade -- Duration of fade-in and fade-out, in seconds (float)
    fs -- Sampling frequency (int)

    Return:
    stimulus -- stereo * 2 signal (4xN numpy array)
    '''

    # Create signal ------------------------------------------------------------
    s = mn.create_sinusoid (freq = frequency, phase = 0, fs = fs, dur = sdur)
    # --------------------------------------------------------------------------

    # Add zeros to signal to make it as long as the white noise ----------------
    s_stimtracker = mn.set_gain(s, -20) # make signal at higher dB without white noice for stimtracker
    s = mn.set_gain(s, s_lvl) # make signal at calibrated lvl
    catch = np.zeros(int(sdur * fs))
    # --------------------------------------------------------------------------

    # Combine noise and signal into stereo -------------------------------------
    fadesamples = int(fade * fs)
    if (condition == 1): # critical
        ch1 = mn.fade(s, samples = fadesamples)
        ch2 = mn.fade(s, samples = fadesamples)
        ch3 = mn.fade(s_stimtracker, samples = fadesamples)
        ch4 = mn.fade(s_stimtracker, samples = fadesamples)
    elif (condition == 2): # catch
        ch1 = mn.fade(catch, samples = fadesamples)
        ch2 = mn.fade(catch, samples = fadesamples)
        ch3 = mn.fade(catch, samples = fadesamples)
        ch4 = mn.fade(s_stimtracker, samples = fadesamples)


    stimulus = np.transpose(np.array([ch1,ch2,ch3,ch4]))
    #stimulus = np.transpose(np.array([ch1,ch2]))
    # --------------------------------------------------------------------------
    return stimulus